/************************************************
* BITLIST.C - routines to manipulate bit lists	*
*												*
* 900406 MCMason - updated:  MSC v5.1			*
* 880624 MCMason - originally coded				*
************************************************/

#include <stdio.h>		/* NULL					*/
#include <stdlib.h>		/* malloc, free			*/
#include <stdarg.h>		/* for variable-args	*/
#include "bitlist.h"

#define  TRUE	(1==1)
#define  FALSE	(1==0)
#define  ListSize(x)	(*(int *)x)
#define  CPI    2		/* # chars in 1 int		*/
#define  CS		8		/* # bits  in 1 char	*/

/************************************************
* blCreate - create a bitlist					*
*												*
* INP:	size - number of bits in the list		*
* OUT:	'handle' of the new bitlist, NULL if	*
*		the bitlist could not be created.		*
************************************************/
BitList blCreate(int size)
	{
	BitList retval;
	retval=(BitList)malloc((size+CS-1)/CS + CPI);
	if (!retval) return (BitList)NULL;
	ListSize(retval)=size;
	blListOp(blCLEAR,retval);
	return retval;
	}

/************************************************
* blDestroy - destroy a bitlist					*
*												*
* INP:	list - the bit list to destroy			*
************************************************/
void blDestroy(BitList list)
	{
	free(list);
	}

/************************************************
* blGetBit - return current bit state			*
*												*
* INP:	bl - pointer to bit list of interest	*
*		BitNum - the bit number of interest		*
* OUT:	FALSE if bit is off or out of range,	*
*		TRUE otherwise.							*
************************************************/
BOOL blGetBit(BitList bl, int BitNum)
	{
	int ByteNum;
	if (BitNum >= ListSize(bl)) return FALSE;
	ByteNum = BitNum/8 + CPI;
	BitNum  = BitNum%8;
	return !!((*bl)[ByteNum] & (0x01<<BitNum));
	}

/************************************************
* blSetBit - set the new bit state				*
*												*
* INP:	bl - pointer to bit list of interest	*
*		BitNum - the bit number of interest		*
*		State - the new bit state				*
* OUT:	TRUE on error, FALSE otherwise			*
************************************************/
BOOL blSetBit(BitList bl, int BitNum, int State)
	{
	int ByteNum,Mask;
	if (BitNum >= ListSize(bl)) return TRUE;
	ByteNum = BitNum/8 + CPI;
	BitNum  = BitNum%8;
	Mask = 0x01<<BitNum;
	if (State)
		(*bl)[ByteNum] |= Mask;
	else
		(*bl)[ByteNum] &= ~Mask;
	return FALSE;
	}

/************************************************
* blListOp - perform a list operation			*
*												*
* INP:	op - operation code to perform:			*
*			 blUNION, blINTERSECT,	blCLEAR,	*
*			 blCLEAR, blSET,		blINVERT	*
*		b1 - bitlist #1							*
*		b2 - bitlist #2 (not required for the	*
*			 ops: blCLEAR, blSET & blINVERT)	*
* OUT:	TRUE if UNION or INTERSECT detect that	*
*		the lists are different sizes, FALSE o/w*
************************************************/
int blListOp(blOpcode op, BitList b1, ...)
	{
	int i, len;
	BitList b2;

	/* Variable-argument code for the optional	*/
	/* second argument.							*/
	va_list am;
	va_start(am,b1);

	/* Be sure that the operations that require */
	/* two lists have lists of matching length.	*/
	if (op==blUNION||op==blINTERSECT||op==blCOPY)
		{
		/* Get second argument if required...	*/
		b2=va_arg(am,BitList);
		if (ListSize(b1) != ListSize(b2))
			return TRUE;
		}

	len = (ListSize(b1)+CS-1)/CS;
	switch (op)
		{
		case blCLEAR:		/* Clear the list	*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i]=0;
				break;

		case blSET:			/* Set the list		*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i]=(char) ~0;
				break;

		case blINVERT:		/* Invert bit list	*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i] = ~(*b1)[i];
				break;

		case blUNION:		/* Merge two lists	*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i] |= (*b2)[i];
				break;

		case blINTERSECT:	/*Get matching parts*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i] &= (*b2)[i];
				break;

		case blCOPY:		/* Copy the list	*/
				for (i=CPI; i<len+CPI; i++)
					(*b1)[i] = (*b2)[i];

		default:			/* Unknown operation*/
				return TRUE;
		}
	return FALSE;
	}
